from pathlib import Path
import typer
import pickledb
import os
import base64

APP_NAME = 'uva-command-line'
CONFIG_FILE_NAME = 'config.db'

DB_KEY_COOKIE = 'COOKIE'
DB_KEY_USERNAME = 'USERNAME'
DB_KEY_UHUNT_UID = 'UHUNT_UID'


def get_database():
    app_dir = typer.get_app_dir(APP_NAME)
    path: Path = Path(app_dir) / CONFIG_FILE_NAME
    if not path.is_file():
        os.makedirs(os.path.dirname(path), exist_ok=True)
    return pickledb.load(path, False)


def read_cookies():
    db = get_database()
    if db.exists(DB_KEY_COOKIE):
        cookies = db.get(DB_KEY_COOKIE)
        return base64.b64decode(cookies)
    else:
        return None


def save_cookies(data):
    encoded = base64.b64encode(data)
    db = get_database()
    db.set(DB_KEY_COOKIE, encoded.decode('ascii'))
    db.dump()


def save_login_data(username, uhunt_uid):
    db = get_database()
    db.set(DB_KEY_USERNAME, username)
    db.set(DB_KEY_UHUNT_UID, uhunt_uid)
    db.dump()


def read_uhunt_uid():
    db = get_database()
    return db.get(DB_KEY_UHUNT_UID)


def purge():
    # app_dir = typer.get_app_dir(APP_NAME)
    # path: Path = Path(app_dir) / CONFIG_FILE_NAME
    # os.remove(os.path.dirname(path))
    db = get_database()
    db.deldb()
    db.dump()



import datetime
from pathlib import Path
from enum import Enum

import timeago


class SubmitLanguage(str, Enum):
    c = 'c',
    java = 'java',
    cplusplus = 'c++'
    pascal = 'pascal'
    cplusplus11 = 'c++11'
    python = 'python'

    def __int__(self):
        if self.value == 'c':
            return 1
        elif self.value == 'java':
            return 2
        elif self.value == 'c++':
            return 3
        elif self.value == 'pascal':
            return 4
        elif self.value == 'c++11':
            return 5
        else:
            return 6


verdict_dict = {
        0: "Processing",
        10: "[bold #000000]Submission error",
        15: "Can't be judged",
        20: "[bold #000000]In queue",
        30: "[bold #AAAA00]Compile error",
        35: "Restricted function",
        40: "[bold #00AAAA]Runtime error",
        45: "Output limit",
        50: "[bold #0000FF]Time limit",
        60: "Memory limit",
        70: "[bold #FF0000]Wrong answer",
        80: "[bold #666600]PresentationE",
        90: "[bold #00AA00]Accepted"
    }

language_dict = {
        1: "ANSI C",
        2: "Java",
        3: "C++",
        4: "Pascal",
        5: "C++11",
        6: "Python"
}

extensions_language_map_dict = {
    '.c': SubmitLanguage.c,
    '.java': SubmitLanguage.java,
    '.cpp': SubmitLanguage.cplusplus,
    '.pas': SubmitLanguage.pascal,
    '.py': SubmitLanguage.python
}


def detect_language(filepath):
    extension = Path(filepath).suffix
    return extensions_language_map_dict.get(extension)


language_extensions_map_dict = {
    SubmitLanguage.c: '.c',
    SubmitLanguage.java: '.java',
    SubmitLanguage.cplusplus: '.cpp',
    SubmitLanguage.pascal: '.pas',
    SubmitLanguage.python: '.py',
}


def get_file_extension(language):
    return language_extensions_map_dict.get(language)


def generate_submission_table_row(s):
    submission_time = timeago.format(datetime.datetime.fromtimestamp(s[4]) - datetime.timedelta())
    return [
        str(s[0]),
        str(s[1]),
        verdict_dict[s[2]],
        str(s[3]),
        submission_time,
        language_dict[s[5]],
        str(s[6])
    ]



from .helpers import detect_language
from .helpers import SubmitLanguage
from .helpers import language_dict
from .helpers import verdict_dict
from .helpers import generate_submission_table_row
from .helpers import get_file_extension



import pickle
import time
import os

import requests
from bs4 import BeautifulSoup
from rich.console import Console
from rich.live import Live
from rich.table import Table

import uva.localdb as localdb
import uva.helpers as helpers

BASE_URL = 'https://onlinejudge.org'
PDF_FILE_URL = BASE_URL + '/external'

LOGIN_PARAMS = {
    'option': 'com_comprofiler',
    'task': 'login',
}

SUBMIT_PARAMS = {
    'option': 'com_onlinejudge',
    'Itemid': 8,
    'page': 'save_submission'
}

UHUNT_BASE_API_URL = 'https://uhunt.onlinejudge.org/api'
UHUNT_UNAME2UID_API_URL = UHUNT_BASE_API_URL + '/uname2uid'
UHUNT_SUBS_USER_API_URL = UHUNT_BASE_API_URL + '/subs-user'
UHUNT_SUBS_USER_LATEST_API_URL = UHUNT_BASE_API_URL + '/subs-user-last'

NOT_AUTHORIEZED_ERROR_STRING = 'You are not authorised to view this resource'
SUBMISSION_SUCESS_MESSAGE = 'mosmsg=Submission+received+with+ID+'

NOT_LOGGED_IN_MESSAGE = "It's seems that you are not logged in, please login first."
SUBMIT_FILE_DOES_NOT_EXIST_OR_EMPTY = "[bold #FF0000]File that you are tying to submit can't be found or it's empty"


def login(username, password):
    console = Console(log_time=False, log_path=False)
    with console.status("[blue]Logging into uva"):
        console.log("Fetching the login form")
        session = requests.Session()
        r = session.get(BASE_URL)
        console.log("Filling out the form")
        soup = BeautifulSoup(r.content, 'html5lib')
        form = soup.find('form', id='mod_loginform')
        inputs = form.find_all('input', type='hidden')

        form_data = {
            'username': username,
            'passwd': password,
            'remember': 'yes'
        }
        for tag in inputs:
            form_data[tag['name']] = tag['value']

        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
        }

        console.log("Submitting the form")
        # TODO add error checks on this call
        response = session.post(BASE_URL, params=LOGIN_PARAMS, headers=headers, data=form_data)

        res = response.content.decode("utf-8")

        if NOT_AUTHORIEZED_ERROR_STRING in res:
            console.log('You are not authorize')
            return
        elif 'My Account' in res and 'Logout' in res:
            console.log('You are logged in')
        else:
            console.log('There was an error')
            return

        console.log("Saving login token to the local db")
        localdb.save_cookies(pickle.dumps(session.cookies))

        console.log("Getting uva hunt username")
        # TODO add error checking on this call
        p = requests.get(UHUNT_UNAME2UID_API_URL + '/' + username)

        console.log("Saving uva hunt username to local db")
        localdb.save_login_data(username, p.content.decode("utf-8"))
        console.log("[bold green]All done")


def get_latest_subs(count):
    console = Console(log_time=False, log_path=False)
    console.log("[blue]Logging into uva")
    cookie = localdb.read_cookies()
    if cookie is None:
        console.log(NOT_LOGGED_IN_MESSAGE)
        return
    console.log('[blue]Getting latest subs')
    uhunt_uid = localdb.read_uhunt_uid()
    url = f'{UHUNT_SUBS_USER_LATEST_API_URL}/{uhunt_uid}/{count}'
    submissions = requests.get(url)
    data = submissions.json()
    console.log(f"[blue]Submissions for user {data['name']}")
    if len(data["subs"]) != 0:
        table = Table(
            "Submission ID", "Problem ID", "Verdict ID", "Runtime", "Submission Time", "Language", "Rank"
        )

        for sub in reversed(data["subs"]):
            table.add_row(*helpers.generate_submission_table_row(sub))

        console.log(table)
        console.log('[blue]All done')
    else:
        console.log("[blue]No submissions for the current user")


def logout():
    console = Console(log_time=False, log_path=False)
    with console.status("[blue]Logging out from uva"):
        localdb.purge()
        console.log("[bold green]Logout success")


def submit(problem_id, filepath, language):
    console = Console(log_time=False, log_path=False)
    console.status("[blue]Submitting your solution")
    cookie = localdb.read_cookies()
    if cookie is None:
        console.log(NOT_LOGGED_IN_MESSAGE)
        return
    session = requests.session()
    session.cookies.update(pickle.loads(cookie))

    if not (os.path.exists(filepath) and os.path.getsize(filepath) > 0):
        console.log(SUBMIT_FILE_DOES_NOT_EXIST_OR_EMPTY)
        return

    files = {
        'localid': (None, problem_id),
        'language': (None, str(language)),
        'codeupl': (filepath, open(filepath, 'rb')),
    }
    console.log("Uploading solution to Uva")
    response = session.post(BASE_URL, params=SUBMIT_PARAMS, files=files)
    res = response.content.decode("utf-8")

    if NOT_AUTHORIEZED_ERROR_STRING in res:
        console.log(NOT_AUTHORIEZED_ERROR_STRING)
    elif SUBMISSION_SUCESS_MESSAGE in res:
        index = res.find(SUBMISSION_SUCESS_MESSAGE)
        end = res.find('"', index)
        submission_id = res[index + len(SUBMISSION_SUCESS_MESSAGE):end]
        console.log(f"Submission with submission id {submission_id} submitted")
        wait_for_submission_results(submission_id, console)
    else:
        console.log('[bold #FF0000]There was an unknow error')


def wait_for_submission_results(submission_id, console=Console(log_time=False, log_path=False)):
    console.log("[bold green]Waiting for results, to exit ctrl + z")
    uhunt_uid = localdb.read_uhunt_uid()
    sub_id = str(int(submission_id) - 1)

    with Live(console=console, auto_refresh=False) as live:
        while True:
            res = requests.get(f"{UHUNT_SUBS_USER_API_URL}/{uhunt_uid}/{sub_id}")

            table = Table(
                "Submission ID", "Problem ID", "Verdict ID", "Runtime", "Submission Time", "Language", "Rank"
            )

            verdict = None
            # TODO this is not good, for some reason uhunt time is ahead for 15 mins.
            if len(res.json()["subs"]) != 0:
                s = res.json()["subs"][0]

                row_data = helpers.generate_submission_table_row(s)
                verdict = s[2]
                table.add_row(*row_data)
            else:
                table.add_row('?', '?', '?', '?', '?', '?', '?')

            live.update(table, refresh=True)

            if verdict is not None and verdict not in [0, 20]:
                break

            time.sleep(3)
    console.log('[bold green]All done')


def get_pdf_file(problem_id):
    # check if this is proper problem id, check if it's inside proper range of problems.
    url = f"{PDF_FILE_URL}/{problem_id[0:3]}/{problem_id}.pdf"
    res = requests.get(url)
    with open(f'{problem_id}.pdf', 'wb') as f:
        f.write(res.content)


def initialize_uva_dir(problem_id, language):
    console = Console(log_time=False, log_path=False)
    console.log(f"[blue]Initializing dir for the problem {problem_id}")
    url = f"{UHUNT_BASE_API_URL}/p/num/{problem_id}"
    problem_data = requests.get(url)
    data = problem_data.json()
    dir_name = f"{problem_id} - {data['title']}"
    console.log(f"[green]Creating directory {dir_name}")
    if os.path.exists(dir_name):
        console.log("[red]Directory already exists")
        return
    os.mkdir(dir_name)
    file_extension = helpers.get_file_extension(language)
    file_name = os.path.join(dir_name, f"main{file_extension}")
    console.log(f"[green]Creating file {file_name}")
    if os.path.exists(file_name):
        console.log("[red]File already exists")
        return
    with open(file_name, 'w') as f:
        f.close()
    console.log(f"All ready to work on the solution for the problem {problem_id}")



import typer
from rich.prompt import Prompt

import uva.commands as commands
import uva.helpers as helpers

app = typer.Typer()

@app.command()
def login(
        username: str = typer.Argument(
            None,
            show_default=False,
            help="Your uva username"
        ),
        password: str = typer.Argument(
            None,
            show_default=False,
            help="Your uva password"
        )
):
    """
        Logs you into the uva portal
    """
    if username is None:
        username = Prompt.ask("Enter your name :sunglasses:")

    if password is None:
        password = Prompt.ask("Enter your uva password", password=True)
    commands.login(username, password)


@app.command()
def logout():
    """
        Logs you out from uva portal
    """
    commands.logout()


@app.command()
def submit(
        problem_id: int = typer.Argument(
            ...,
            show_default=False,
            help="Uva problem id"
        ),
        filepath: str = typer.Argument(
            ...,
            show_default=False,
            help="Path to the solution file"
        ),
        language: helpers.SubmitLanguage = typer.Argument(
            None,
            show_default=False,
            help="Submission language"
        )
):
    """
        Submits the problem to uva portal
    """
    if language is None:
        language = helpers.detect_language(filepath)

    if language is None:
        ask_user = Prompt.ask("Please enter submit language", choices=['c', 'java', 'c++', 'pascal', 'c++11', 'python'])
        language = helpers.SubmitLanguage(ask_user)
    elif language is helpers.SubmitLanguage.cplusplus:
        ask_user = Prompt.ask("Please enter c++ version", choices=['c++', 'c++11'], default='c++')
        language = helpers.SubmitLanguage(ask_user)
    commands.submit(problem_id, filepath, language)


@app.command()
def latest_subs(count: int = typer.Argument(
        10,
        min=0,
        max=100,
        help='Number of submissions to get'
    )
):
    """
        Prints latest subs
    """
    commands.get_latest_subs(count)


@app.command()
def pdf(problem_id: int = typer.Argument(
        ...,
        show_default=False,
        help="Uva problem id"
    )
):
    """
        Downloads problem pdf
    """
    commands.get_pdf_file(str(problem_id))


@app.command()
def start(
    problem_id: int = typer.Argument(
        ...,
        show_default=False,
        help="Uva problem id"
    ),
    language: helpers.SubmitLanguage = typer.Argument(
        None,
        show_default=False,
        help="Submission language"
    )
):
    """
        Initialize directory to work on the problem
    """
    if language is None:
        ask_user = Prompt.ask("Please enter submit language", choices=['c', 'java', 'c++', 'pascal', 'c++11', 'python'])
        language = helpers.SubmitLanguage(ask_user)
    commands.initialize_uva_dir(str(problem_id), int(language))

# @app.command()
# def test(subid: str):
#     console = Console(log_time=False, log_path=False)
#     commands.wait_for_submission_results(subid, console)


if __name__ == "__main__":
    app()
